/*---------------------------------------------------------------------------*/
/*      FFF.COM - Fast-File-Find, a fully public domain program to search
 * 				  for files on one or more disks (functionally similar to
 * 				  the familiar "WHERE" program.  FFF will also search inside
 * 				  of ZIP, ARC, PKA, and ZOO archives to the first level.
 *
 *      Version:  2.1.1
 *      Date:     November 15, 1989
 *      Author:   Don A. Williams
 *
            *********************  NOTICE  ************************
            *  Contrary to the current trend in MS-DOS  software  *
            *  this  program,  for whatever it is worth,  is NOT  *
            *  copyrighted (with the exception  of  the  runtime  *
            *  library  from  the C compiler)!  The program,  in  *
            *  whole or in part,  may  be  used  freely  in  any  *
            *  fashion or environment desired.  If you find this  *
            *  program  to  be  useful  to you,  do NOT send any  *
            *  contribution to the author;  in the words of Rick  *
            *  Conn,   'Enjoy!'   However,   if   you  make  any  *
            *  improvements,  I would enjoy receiving a copy  of  *
            *  the  modified source.  I can be reached,  usually  *
            *  within 24  hours,  by  messages  on  any  of  the  *
            *  following Phoenix, AZ systems:                     *
            *                                                     *
            *     The Tool Shop BBS       [PCBOARD] [PC-Pursuit]  *
            *         (602) 279-2673   1200/2400/9600 bps         *
            *     Technoids Anonymous     [PCBOARD]               *
            *         (602) 899-4876   300/1200/2400 bps          *
            *         (602) 899-5233                              *
            *         (602) 786-9131                              *
            *     Inn On The Park         [PCBOARD]               *
            *         (602) 957-0631   1200/2400/9600 bps         *
            *     Pascalaholics Anonymous [WBBS]                  *
            *         (602) 484-9356   1200/2400 bps              *
            *                                                     *
            *  or:                                                *
            *     GEnie, mail address: DON-WILL                   *
            *     CompuServ:           75410,543                  *
            *                                                     *
            *  Every  effort  has  been  made to avoid error and  *
            *  moderately extensive testing has  been  performed  *
            *  on  this  program,  however,  the author does not  *
            *  warrant it to be fit for any  purpose  or  to  be  *
            *  free  from  error and disclaims any liability for  *
            *  actual or any other damage arising from  the  use  *
            *  of this program.                                   *
            *******************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <dir.h>
#include <io.h>

#define MAIN

#define VERSION "Version 2.1.1"
#define VER_DATE "November 15, 1989"

#include "queue.h"
#include "arc.h"
#include "zip.h"
#include "zoo.h"

#ifndef DOS_DATE
#define DOS_DATE
typedef union {
	unsigned u;
	struct {
		unsigned Day : 5;
		unsigned Month : 4;
		unsigned Year  : 7;
		} b;
	} DOS_FILE_DATE;
#endif

#ifndef DOS_TIME
#define DOS_TIME
typedef union {
	unsigned u;
	struct {
		unsigned Second : 5;
		unsigned Minute : 6;
		unsigned Hour   : 5;
		} b;
	} DOS_FILE_TIME;
#endif

void WalkTree (void);
int Match (char *Str, char *Pat);
void GetPath (QUE_DEF *Q, char *Arg);
void PrtVerbose (char *Path, char *Name, DOS_FILE_TIME Time, DOS_FILE_DATE Date,
				 long Size);
void Usage (void);
void ChkPage (void);

void GetCurrentDisk (int *CurDisk);
int GetCurrentDirectory (int Disk, char *CurDir);
void ErrorExit (char *Format, ...);
void GetProgName (char *Name, char *argv);

char Path[65];
char Temp[66];
char Pattern[66];
char CurDir[65];
unsigned long Position;
char ProgName[9];
int Spaced     = 0;
int Verbose    = 0;
int Page       = 0;
int Lnno       = 1;
int TotalFiles = 0;
int TotalMatch = 0;
int ArcFiles   = 0;
int ArcedFiles = 0;
int ArcMatch   = 0;
int ZipFiles   = 0;
int ZipedFiles = 0;
int ZipMatch   = 0;
int ZooFiles   = 0;
int ZooedFiles = 0;
int ZooMatch   = 0;

int ArcSwt		= 1;


 void
main (int argc, char *argv[]) {
	QUE_DEF Paths;
	QUE_ENTRY *t;
	int CurDisk, i;
	char *p;

	GetProgName(ProgName, argv[0]);
	fprintf(stderr, "%s: FastFileFind - %s: %s\n", ProgName, VERSION, VER_DATE);
	InitQueue(&Paths);
	GetCurrentDisk(&CurDisk);
	CurDir[0] = (char) (CurDisk + 'A'); CurDir[1] = ':'; CurDir[2] = '\\';
	GetCurrentDirectory(0, CurDir+3);
	strupr(CurDir);
	strcpy(Path, "C:\\");
	if (argc < 2) Usage();
	for (i=1; i < argc; ++i) {
		if (argv[i][0] == '-') {
			switch( tolower(argv[i][1]) ) {
				case 'a':
					ArcSwt = 0;
					break;
				case 'v':
					Verbose = 1;
					break;
				case 'p':
					Page = 1;
					break;
				default:
					fprintf(stderr, "Invalid argument: %c\n", argv[i][1]);
					Usage();
					break;
				}
			continue;
			}
		else if ( (p = strrchr(argv[i], '\\')) != NULL ) {
			strcpy(Pattern, p+1);
			*(p+1) = '\0';
			}
		else if ( (p = strchr(argv[i], ':')) != NULL ) {
			strcpy(Pattern , p+1);
			*(p+1) = '\0';
			}
		else {
			strcpy(Pattern, argv[i]);
			Path[0] = (char) (CurDisk + 'A'); Path[1] = ':'; Path[2] = '\\';
			Path[3] = '\0';
			Enque(&Paths, Path);
			continue;
			}
		GetPath(&Paths, argv[i]);
		}
	strupr(Pattern);
	for (t = Paths.Head; t != NULL; t = t->Next) {
		strcpy(Path, t->Body);
		WalkTree();
		}
	if (!Spaced) printf("\n");
	printf("Total Files = %6d  Total Matched Files = %d\n", TotalFiles, TotalMatch);
    printf("ARC Files   = %6d  ", ArcFiles);
    if (ArcSwt) printf("ARCed Files = %6d  ARC Matches = %d", ArcedFiles, ArcMatch);
    printf("\nZIP Files   = %6d  ", ZipFiles);
    if (ArcSwt) printf("ZIPed Files = %6d  ZIP Matches = %d", ZipedFiles, ZipMatch);
    printf("\nZOO Files   = %6d  ", ZooFiles);
    if (ArcSwt) printf("ZOOed Files = %6d  ZOO Matches = %d", ZooedFiles, ZooMatch);
    printf("\n");
	}

 void
WalkTree (void) {
	extern int Verbose, Spaced;

	int Status, Reply;
	QUE_DEF Direc;
	QUE_ENTRY *t, *u;
	char *p;
	struct ffblk DirBlk;
	DOS_FILE_TIME Time;
	DOS_FILE_DATE Date;

	InitQueue(&Direc);
	strcat(Path, "*.*");
	Status = findfirst(Path, &DirBlk, 0xFF);
	*(strrchr(Path, '\\')+1) = '\0';
	while (!Status) {
		if ( (DirBlk.ff_attrib & FA_DIREC) != 0) {
			if (DirBlk.ff_name[0] != '.') {
				Enque(&Direc, DirBlk.ff_name);
				}
			}
		else {
			++TotalFiles;
			if ( Match(DirBlk.ff_name, Pattern) ) {
				++TotalMatch;
				if (Page) ChkPage();
				if (Verbose) {
					fputs("  ", stdout);
					Time.u = DirBlk.ff_ftime;
					Date.u = DirBlk.ff_fdate;
					PrtVerbose(Path, DirBlk.ff_name, Time, Date,
							   DirBlk.ff_fsize);
					}
				else {
					fputs(Path, stdout);
					puts(DirBlk.ff_name);
					}
				++Lnno;
				Spaced = 0;
				}
			if ( (p = strchr(DirBlk.ff_name, '.')) != NULL ) {
				if ( (stricmp(p+1, "ARC") == 0) || (stricmp(p+1, "PKA") == 0) ) {
					++ArcFiles;
					if (ArcSwt) {
						strcat(Path, DirBlk.ff_name);
						DoArc(Path);
						*(strrchr(Path, '\\') + 1) = '\0';
						}
					}
				else if (stricmp(p+1, "ZIP") == 0) {
					++ZipFiles;
					if (ArcSwt) {
						strcat(Path, DirBlk.ff_name);
						DoZip(Path);
						*(strrchr(Path, '\\') + 1) = '\0';
						}
					}
				else if (stricmp(p+1, "ZOO") == 0) {
					++ZooFiles;
					if (ArcSwt) {
						strcat(Path, DirBlk.ff_name);
						DoZOO(Path);
						*(strrchr(Path, '\\') + 1) = '\0';
						}
					}
				}
			}
		Status = findnext(&DirBlk);
		}
	p = strrchr(Path,'\\') + 1;
	for (t=Direc.Head; t != NULL; ) {
		*p = '\0';
		strcat(Path, t->Body); strcat(Path, "\\");
		WalkTree();
		u = t->Next;
		free(t->Body);
		free(t);
		t = u;
		}
	}

 int
Match (char *Str, char *Pat) {

	while (*Str != '\0') {
		if (*Pat == '*') {
			if ( *(++Pat) == '\0') return(1);
			else {
				while (*Str != *Pat && *Str != '\0') ++Str;
				if (*Str == '\0') {
					if (*Pat == '.' && ( *(Pat+1) == '\0' || *(Pat+1) == '*') ) return(1);
					else return(0);
					}
				}
			}
		if (*Pat == '?') {
			++Pat; ++Str;
			}
		if (*Str != *Pat) return(0);
		++Pat; ++Str;
		}
	if ( (*Pat == '\0') || !strcmp(Pat, ".*") || !strcmp(Pat, "*") ) return(1);
	else return(0);
	}


 void
GetPath (QUE_DEF *Q, char *Arg) {
	char *p, Temp[65];
	int i, j;

	if ( (p = strchr(Arg, ':')) != NULL) {
		if ( (i = p - Arg) > 1) {
			for (j=0; j < i; ++j) {
				Temp[0] = Arg[j];
				Temp[1] = '\0';
				strcat(Temp, p);
				if (Temp[strlen(Temp)-1] != '\\') strcat(Temp, "\\");
				strupr(Temp);
				if (Enque(Q, Temp) == NULL)
					ErrorExit("Insufficient memory for Path List.\n");
				}
			}
		else {
			strcpy(Temp, Arg);
			if (Temp[strlen(Temp)-1] != '\\') strcat(Temp, "\\");
			strupr(Temp);
			if (Enque(Q, Temp) == NULL)
				ErrorExit("Insufficient memory for Path List.\n");
			}
		}
	else {
		Temp[0] = CurDir[0]; Temp[1] = ':'; Temp[2] = '\0';
		strcat(Temp, Arg);
		if (Temp[strlen(Temp)-1] != '\\') strcat(Temp, "\\");
		strupr(Temp);
		if (Enque(Q, Temp) == NULL)
			ErrorExit("Insufficient memory for Path List.\n");
		}
	}

 void
ErrorExit (char *Format, ...) {
	va_list ArgPtr;
	extern char ProgName[];
	extern int errno;

	fprintf(stderr, "%s: ", ProgName);
	va_start(ArgPtr, Format);
	vfprintf(stderr, Format, ArgPtr);
	va_end(ArgPtr);
	if (errno) perror("    ");
	fprintf(stderr, "\n");
	exit(1);
	}

 void
GetProgName (char *Name, char *argv) {
	char *p, *p1;

	if ( (p = strrchr(argv, '\\')) != NULL) ++p;
	else p = argv;
	if ( (p1 = strchr(p, '.')) != NULL) *p1 = '\0';
	strcpy(Name, p);
	}

 void
PrtVerbose (char *Path, char *Name, DOS_FILE_TIME Time, DOS_FILE_DATE Date,
		    long Size) {
	static char *MonTab[] = {
		"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
		"Oct", "Nov", "Dec"
		};

	printf("%-13s   %02u %s %02u   %02u:%02u %8ld   %s\n", Name, Date.b.Day,
		   MonTab[Date.b.Month - 1], Date.b.Year + 80, Time.b.Hour,
		   Time.b.Minute, Size, Path);
	}

 void
Usage (void) {
	fprintf(stderr, "USAGE: %s [device(s):[path]pattern [options]\n", ProgName);
	fprintf(stderr, "\nOptions:\n\n");
	fprintf(stderr, "    -a Suppress searching inside of archive files.\n");
	fprintf(stderr, "    -v Display Date/Time and File Size as well as name\n");
	fprintf(stderr, "       and path for all matched files.\n");
	fprintf(stderr, "    -p Paginate output every 23 lines\n");
	exit(1);
	}


 void
ChkPage (void) {
	int Reply;

	if (Lnno >= 24) {
		clreol();
		printf("More?...");
		Reply = tolower( getche() );
		if ( (Reply == 'n') || (Reply == 0x03) ) exit(1);
		printf("\r"); clreol();
		Lnno = 1;
		}
	}